class OSIM_RL:
    PING = "OSIM_RL.PING"
    PONG = "OSIM_RL.PONG"
    ENV_CREATE = "OSIM_RL.ENV_CREATE"
    ENV_CREATE_RESPONSE = "OSIM_RL.ENV_CREATE_RESPONSE"
    ENV_RESET = "OSIM_RL.ENV_RESET"
    ENV_RESET_RESPONSE = "OSIM_RL.ENV_RESET_RESPONSE"
    ENV_STEP = "OSIM_RL.ENV_STEP"
    ENV_STEP_RESPONSE = "OSIM_RL.ENV_STEP_RESPONSE"
    ENV_SUBMIT = "OSIM_RL.ENV_SUBMIT"
    ENV_SUBMIT_RESPONSE = "OSIM_RL.ENV_SUBMIT_RESPONSE"
    ERROR = "OSIM_RL.ERROR"
